/* in this module appeaars machinery connected with the recognition
   storage of words */

#include "align.h"     /* align definitions */
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>


/* externals from errors.c */
extern goof(); /* error-writing routine
  void goof(char *s); */
extern int sinned, error; /* error-flags */

/* externals from slurp.c */
extern nextchar(); /* gets next character
  char nextchar(); */
extern peekchar(); /* peeks at next char
  char peekchar() */
extern advance();  /* advances to next printing char
  void advance(); */

/* externals from glosses */
extern char delmark;     /* example delimiter */
extern int wordmax;      /* maximum length of word */

/* word storage */

static char store[charmax];  /* array for storing string */
static char *free_char;      /* pointer to first available position */
static char *top_char;       /* pointer to top of store */

/* initializes store */
init_store()
{
  free_char = store;
  top_char = store+charmax;
}

/* stows characters */
int stow(c)
char c;
{
  if (free_char<top_char) {
    *(free_char++) = c;
    return(TRUE);
  }
  else {
    goof("too many characters in example");
  }
}

/*   Scribe employs several kinds of brackets (delimiters, in Scribe
     jargon, and in macros they can't be nested freely.  So some fairly
     complicated machinery is required to keep track of them.  Align
     assumes that anything bounded by command-delimiters (e.g.
     `@i< ... >' is part of a single word  */

/* define bracket types & sides */
#define parens 0
#define curlies 6
#define LEFT 0
#define RIGHT 1

char brackets[7][2] = {
  {'(',')'},
  {'[',']'},
  {'\'','\''},
  {'\`','\''},
  {'"','"'},
  {'<','>'},
  {'{','}'}
};

/* resets a list of bracket-types (reg) to say that all are available
   (e.g. haven't been used inside the thing you're trying to put
   brackets around */
void clear_brackets(reg)
int reg[curlies];
{
  int i;

  for (i=parens; i<=curlies; i++)
  {
    reg[i]=TRUE;
  }
}

/* says whether `c' is a bracket of side `s', and puts its type into *t */
int bracket_type(c,s,t)
char c;
int s;
int *t;
{
  int i;

  for (i = parens; i <= curlies; i++) {
    if (brackets[i][s] == c) {
      *t = i;
      return(TRUE);
    }
  }
  return(FALSE);
}

/* return brackets from codes */
char right_of(d)
int d;
{return(brackets[d][RIGHT]);}

char left_of(d)
int d;
{return(brackets[d][LEFT]);}

/* Takes a register of bracket-types (reg), and returns TRUE
   if any are useable, indicating type in `br' */
int useable_bracket(reg,br)
int reg[];
int *br;
{
  int i;
  for (i=parens; i<=curlies; i++) {
    if (reg[i] == TRUE) {
      *br = i;
      return(TRUE);
    }
  }
  return(FALSE);
}


/* machinery for dealing with Scribe Commands */

/* picks out characters that can appear in commands that take arguments
   (i.e., all but `@ ' and `@#', the two sapce commands */
#define iscomchar(c) (isalnum(c) || c=='+' || c=='-')

/* various kinds of scribe operators */
#define scribeops0(c) (c == '\\' || c == '^' || c == '/' || c== '!')
#define scribeops1(c) (c == ':' || c == ' ' || c == '@')
#define scribeops(c) (c == '#' || scribeops0(c) || scribeops1(c))

/* structure for storing width information */
struct comwidth {
  char name[6];
  int length;
  int type;
  struct comwidth *next;
};

typedef struct comwidth com_width;

/* variable for storing list of char_com info */
static com_width *com_widths = NULL;

/* stacks for brackets and whether overprinting is in effect */
static char delstack[stackmax];  /* brackets */
static int  ovpstack[stackmax];  /* overprinting or not */
static int  envpointer;            /* pointer for these */

/* Procedure for reading in char_com info.  Filename passed as string */
int get_com_widths(fname)
char *fname;     /* name of com width file */
{

  FILE *f;         /* file pointer */
  char t[6];       /* store for true/false word infile */
  com_width *cw, *cc; /* pointer for new & current com_width record holder */


  if ((f = fopen(fname,"r")) == NULL)
    return(0);
  cc=NULL;
  cw = (struct comwidth *) malloc(sizeof(struct comwidth));
  while (fscanf(f,"%s%d%s",cw->name,&(cw->length),t) != EOF)  {
    if (t[0]=='t') (cw->type)=TRUE;
    else
      (cw->type)=FALSE;
    (cw->next)=cc;
    cc=cw;
    cw = (com_width *) malloc(sizeof(com_width));
  };
  com_widths=cc;
  fclose(f);
  free(cw);
  return(1);
}

/* looks up command labl, returning pointer to its com_width record, if
   any; otherwise NULL */
com_width *has_com_width(labl)
char      labl[];
{

  com_width *com;

  com = com_widths;
  while (com != NULL) {
    if (!(strcmp(com->name,labl))) /* strcmp returns 0=false if str =) */
      break;
    else
      com = (com->next);
  }
  return(com);
}

/* pushes right delimiter onto env stack, and updates ovpstack also */
push_env(c)
char c;
{
  envpointer++;
  if (envpointer>stackmax) {
    goof("environments nested too deeply in word");
    return;
  }
  if (envpointer == 1)
    ovpstack[envpointer] = 0; /* overprinting is initially off */
  else
    ovpstack[envpointer] = ovpstack[envpointer-1];  /* but inherited from
						     containing env. */
  delstack[envpointer]=c;
}


/* reads in scribe commands */
void get_com(com,grab,stash)
char *com;        /* pointer to some storage for a copy of the command */
char (*grab)();   /* function for reading characters */
void (*stash)();  /* function for stashing characters */
{
  char c;
  int b;

  c=(*grab)();
  (*stash)(c);
  *com++=c;
  if iscomchar(c) {
    do {
      c=(*grab)();
      *com++=c;
      (*stash)(c);
    }
    while iscomchar(c);
    *(--com) = '\0';
    if (bracket_type(c,LEFT,&b)) {
      push_env(right_of(b));
    }
    else
      goof("scribe command not followed by left delimiter");
  }
  else
    *com = '\0';
}

/* returns whether overprinting is current */
int overprint()
{
  if (envpointer == 0)
    return(FALSE);
  else return(ovpstack[envpointer]);
}

#define shortie(c) (c == 'I' ||  c== 'J' || c == 'S')
#define longie(c) ((isupper(c) && !shortie(c)) || c == 'm' || c == 'w')

void readword(here, stop, d, grab, lookahead)
word *here;      /* location of word record to be filled in */
int *stop;       /* where to record whether stop-delimiter encountered */
char d;          /* a parametric word-ending character */
char (*grab)();  /* character-reading procedure to be used */
int lookahead;   /* whether to advance to next printing char */
{

  char c;                 /* character holder */
  int cc;                 /* counts capitals */
  int brackets[curlies];  /* bracket-register */
  char com[20];           /* command-holder */
  com_width *com_ptr;     /* pointer to command record */
  int size;               /* size of word */
  int bracket;            /* bracket-type */

  *stop = FALSE;
  envpointer = 0;
  cc = 0;
  clear_brackets(brackets);

  here->loc=free_char;
  size=0;
  c=grab();
  while (envpointer > 0 || !(c == d || c == delmark)) {  /* there is nothing
      in the stack, or neither of the word-enders encountered */
    stow(c);
    if (error) return;
    else
    if (c == '@') {
      get_com(com,grab,stow);
      if (error) return;
      if scribeops1(com[0]) size++;
      else
      if (com[0] == '#') size = size+2;
      else
      if ((com_ptr = has_com_width(com)) != NULL) {
	if (com_ptr->length == 0) ovpstack[envpointer]=TRUE;
	else {
	  size = size+(com_ptr->length);
	  if (com_ptr->type == 1) cc++;
	}
      }
    }
    else
    if (envpointer > 0 && c == delstack[envpointer])
      envpointer--;
    else
    if (!overprint()) {
      size++;
      if longie(c) {
       cc++;
      }
    }
    if (bracket_type(c,RIGHT,&bracket)) {
      brackets[bracket] = FALSE;
    }
    c=grab();
    if (c == delmark && envpointer > 0) {
      goof("line-break in Scribe environment");
      envpointer = 0;
    }
  }

  stow('\0'); /* mark end of word i store */
  size = size + (int) floor(((double) cc)/2);
  here->size = size;
  if (c == delmark) *stop = TRUE;
  if (!*stop) {
    if (lookahead) advance();
    if (peekchar()==delmark) {
      *stop = TRUE;
      nextchar();
    }
  }
  if (here->size>wordmax) goof("warning: possible overlong word");
  if (!useable_bracket(brackets,&(here->bracket)))
    goof("no available delimiter for word");
}

/* yet another version of readword ... */

void readword2(here, stop, quit, grab, lookahead)
word *here;      /* location of word record to be filled in */
int *stop;       /* where to record whether stop-delimiter encountered */
int (*quit)();          /* a quit function */
char (*grab)();  /* character-reading procedure to be used */
int lookahead;   /* whether to advance to next printing char */
{

  char c;                 /* character holder */
  int cc;                 /* counts capitals */
  int brackets[curlies];  /* bracket-register */
  char com[20];           /* command-holder */
  com_width *com_ptr;     /* pointer to command record */
  int size;               /* size of word */
  int bracket;            /* bracket-type */

  *stop = FALSE;
  envpointer = 0;
  cc = 0;
  clear_brackets(brackets);

  here->loc=free_char;
  size=0;
  c=grab();
  while (envpointer > 0 || !quit(c)) {  /* there is nothing
      in the stack, or neither of the word-enders encountered */
    stow(c);
    if (error) return;
    else
    if (c == '@') {
      get_com(com,grab,stow);
      if (error) return;
      if scribeops1(com[0]) size++;
      else
      if (com[0] == '#') size = size+2;
      else
      if ((com_ptr = has_com_width(com)) != NULL) {
	if (com_ptr->length == 0) ovpstack[envpointer]=TRUE;
	else {
	  size = size+(com_ptr->length);
	  if (com_ptr->type == 1) cc++;
	}
      }
    }
    else
    if (envpointer > 0 && c == delstack[envpointer])
      envpointer--;
    else
    if (!overprint()) {
      size++;
      if longie(c) {
       cc++;
      }
    }
    if (bracket_type(c,RIGHT,&bracket)) {
      brackets[bracket] = FALSE;
    }
    c=grab();
    if (c == delmark && envpointer > 0) {
      goof("line-break in Scribe environment");
      envpointer = 0;
    }
  }

  stow('\0'); /* mark end of word i store */
  size = size + (int) floor(((double) cc)/2);
  here->size = size;
  if (c == delmark) *stop = TRUE;
  if (!*stop) {
    if (lookahead) advance();
    if (peekchar()==delmark) {
      *stop = TRUE;
      nextchar();
    }
  }
  if (here->size>wordmax) goof("overlong word");
  if (!useable_bracket(brackets,&(here->bracket)))
    goof("no available delimiter for word");
}

/* readword hacked for the peculiar demands of the `columns'
module, essentially being that length calculations are done so
that alignments can be worked out on the screen */

void readword_col(here, bar, quit, grab)
word *here;      /* location of word record to be filled in */
int *bar;       /* where to record whether bar-delimiter encountered */
int (*quit)();          /* function specifying whether to stop */
char (*grab)();  /* character-reading procedure to be used */
{

  char c;                 /* character holder */
  int brackets[curlies];  /* bracket-register */
  char com[20];           /* command-holder */
  com_width *com_ptr;     /* pointer to command record */
  int size;               /* size of word */
  int bracket;            /* bracket-type */

  *bar = FALSE;
  envpointer = 0;
  clear_brackets(brackets);

  here->loc=free_char;
  size=0;
  c=grab();
  while (envpointer > 0 || !quit(c)) {  /* there is nothing
      in the stack, or neither of the word-enders encountered */
    stow(c);
    if (error) return;
    else
    if (c == '@') {
      get_com(com,grab,stow);
      if (error) return;
      if scribeops1(com[0]) size++;
      else
      if (com[0] == '#') size = size+2;
      else
      if ((com_ptr = has_com_width(com)) != NULL) {
	if (com_ptr->length == 0) ovpstack[envpointer]=TRUE;
	else {
	  size++;
	}
      }
    }
    else
    if (envpointer > 0 && c == delstack[envpointer])
      envpointer--;
    else
    if (!overprint()) {
      size++;
    }
    if (bracket_type(c,RIGHT,&bracket)) {
      brackets[bracket] = FALSE;
    }
    c=grab();
    if (eoln() && envpointer > 0) {
      goof("line-break in Scribe environment");
      envpointer = 0;
    }
  }

  stow('\0'); /* mark end of word i store */
  here->size = size;
  if (c == '|') *bar = TRUE;

  if (!useable_bracket(brackets,&(here->bracket)))
    goof("no available delimiter for word");
}

/* writes words to a stream */
void writeword(out,it)
FILE *out;
word *it;
{
  fprintf(out,"%c%s%c",brackets[it->bracket][LEFT],
		       it->loc,
		       brackets[it->bracket][RIGHT]);
}

/*
char delmark = '\\';
int  wordmax = 60;

main()
{
word sample;
int stop;

get_com_widths("widths.dat");
open_source("text.dat");
init_store();


advance();
readword(&sample,&stop,' ',nextchar,0);
printf("%s, size = %d, bracket= %d\n",sample.loc,sample.size,sample.bracket);
if (stop) printf("stop!\n");

}

*/
